/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLEnum.h"
#import <Foundation/Foundation.h>
#import "IDLDefinition_Private.h"
#import "IDLTypedValue.h"
#import "IDLDefinitionScope_Private.h"

@implementation IDLEnum

static int _cmpEnum(id key0, id key1, id dict);

- (void) _resetNames
{
    [_strings release];
    _strings = [[[_enumNames allKeys] sortedArrayUsingFunction:_cmpEnum context:_enumNames]retain];
}

- (void) _calcValue
{
    NSEnumerator* en = [_enumNames objectEnumerator];
    NSNumber* v0 = [en nextObject];
    NSNumber* v;
    while(v = [en nextObject])
        if([v0 compare:v] == NSOrderedDescending)
            v0 = v;
    _value = [v0 copy];
}

- (void) _registerEntries
{
    NSEnumerator* en = [_enumNames keyEnumerator];
    NSString* string;
    while(string = [en nextObject]){
        IDLTypedValue* obj = [IDLTypedValue valueOfType:self withValue:string];
        [obj setVisible:NO];
        [[IDLDefinitionScope currentScope]addDefinition:obj withName:string];
    }
}

- initWithDictionary:(NSDictionary*)dict name:(NSString*)aName
{
    self = [super init];
    _enumNames = [dict retain];
    _name = [aName copy];
    [self _calcValue];
    [self _registerEntries];
    [self _resetNames];
    return self;
}

+ enumWithDictionary:(NSDictionary*)dict name:(NSString*)aName
{
    return [[[self allocWithZone:[self zone]]initWithDictionary:dict name:aName]autorelease];
}

+ enumWithDictionary:(NSDictionary*)dict
{
    return [self enumWithDictionary:dict name:nil];
}

- (void) dealloc
{
	[_enumNames release];
	[_name release];
	[_strings release];
	[_value release];
	[super dealloc];
}

- (NSString*)name
{
    if(_name) 
		return _name;
    else 
		return [super name];
}

- (NSString*) description
{
    return [[super description] stringByAppendingFormat:@" enum %@",[self name]];
}


static int _cmpEnum(id key0, id key1, id dict)
{
	id obj0 = [dict objectForKey:key0];
	id obj1 = [dict objectForKey:key1];
	if(obj0 && obj1)
		return [(NSString*)obj0 compare:obj1];
	else
		return obj0?(obj1?NSOrderedAscending:NSOrderedSame):NSOrderedDescending;
}

- valueFromString:(NSString*)string
{
    return [_enumNames objectForKey:string]; 
}
- (NSNumber*) enumValueForKey:(NSString*)key // raises exception
{
    NSNumber* n = [_enumNames objectForKey:key];
    if(n)
        return n;
    else
        [NSException raise:IDLInvalidArgumentException
            format:@"Enumeration '%@' does not have '%@' value",
            [self name],key];
    return nil; //for compiler
}

- valueForKey:(NSString*)key
{
	return [_enumNames objectForKey:key];
}

// string --> number
- convertValue:aValue
{
    if([aValue isKindOfClass:[NSString class]])
        return [self valueFromString:(NSString*)aValue];
    else
    if([aValue isKindOfClass:[NSNumber class]]){
        const char* valueType = [(NSNumber*)aValue objCType];
        switch(valueType[0]){
            case 'c':
            case 's':
            case 'i':
            case 'l':
            case 'q':
            case 'C':
            case 'S':
            case 'I':
            case 'L':
            case 'Q':
                return aValue;
            default:
                return nil;
        }
    }else
        return nil;
}

//number -->string
- _coerceValue:aValue
{
    if([aValue isKindOfClass:[NSString class]]){
        if([_enumNames objectForKey:(NSString*)aValue])
            return aValue;
        else
            return nil;
    }else
    if([aValue isKindOfClass:[NSNumber class]]){
        return [self stringFromValue:aValue];
    }else
        return nil;
}

//number -->string
- coerceValue:aValue
{
    if(aValue){
        id obj = [self _coerceValue:aValue];
        if(obj)
            return obj;
        else{
            [NSException raise:@"InvalidEnumValueException"
                format:@"invalid choice '%@' for 'enum %@'",
                [aValue description], _name];
            return nil;
        }
    }else
        return nil;
    
}

- (NSString*) valueDescriptionOf:value
{
    return [self coerceValue:value];
}

- (NSString*) stringFromValue:(NSNumber*)value
{
    NSEnumerator* en = [_enumNames keyEnumerator];
    NSString* string;
    while(string = [en nextObject]){
        if([value isEqual:[_enumNames objectForKey:string]])
            break;
    }
    return string;
}

- (long) kind
{
    return tk_enum;
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	[super encodeWithCoder:coder];
	if([coder allowsKeyedCoding]){
		[coder encodeObject:_enumNames forKey:@"enum"];
		[coder encodeObject:_name forKey:@"name"];
	}else{
		[coder encodeObject:_enumNames];
		[coder encodeObject:_name];
	}
}

- initWithCoder:(NSCoder*)coder
{
	self = [super initWithCoder:coder];
	if([coder allowsKeyedCoding]){
		_enumNames = [[coder decodeObjectForKey:@"enum"]retain];
		_name = [[coder decodeObjectForKey:@"name"]retain];
	}else{
		_enumNames = [[coder decodeObject]retain];
		_name = [[coder decodeObject]retain];
	}
	return self;
}

- (id)awakeAfterUsingCoder:(NSCoder *)aDecoder
{
	id obj = [(NSObject*)[[IDLDefinitionScope currentScope]definitionForTypeId:[self typeId]]retain];
	if(obj && obj != self){
		[self release];
		return obj;
	}else{
		[self _calcValue];
		[self _registerEntries];
		[self _resetNames];
		return self;
	}
}

@end
