/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "IDLDefinition_Private.h"
#import "IDLInterface_Private.h"
#import "IDLFunction_Private.h"
#import "IDLAttribute_Private.h"
#import "IDLDefinitionScope.h"
#import "IDLArray.h"
#import "IDLArrayIdentifier.h"
#import "IDLTypedef.h"
#import "IDLEnum.h"
#import "IDLAbstractBase.h"

@implementation IDLInterface

extern NSString* currentParserFile();
static NSLock* _lock = nil;
static IDLInterface* _rootInterface = nil;

+ (void) becomeMultiThreaded:(NSNotification*)notification
{
    _lock = [[NSLock alloc]init];
}

+ (void) initialize
{
    if([IDLInterface class] == self){
        if([NSThread isMultiThreaded])
            [self becomeMultiThreaded:nil];
        else
            [[NSNotificationCenter defaultCenter] addObserver:self
                selector:@selector(becomeMultiThreaded:)
                name:NSWillBecomeMultiThreadedNotification object:nil];
    }
}

+ (IDLInterface*) rootInterface
{
    if(!_rootInterface)
    	_rootInterface = (IDLInterface*)[[IDLDefinitionScope rootScope]definitionForNamePath:@"CORBA/Object"];
    return _rootInterface;
}        

- (void) setName:(NSString*)aName
{
    if(_name != aName){
        [_name release];
	_name = [aName copy];
    }
}

- (void) setInheritanceList:(NSArray*)aList
{
    if(!_inheritance)
	_inheritance = [aList retain];
}

- (void) setFunctionList:(NSDictionary*)aList
{
    if(!_list)
	_list = [aList mutableCopy];
}

- (NSArray*) inheritanceList
{
    return _inheritance;
}

- (BOOL) isAbstract
{
    return _isAbstract;
}

- (void) setIsAbstract:(BOOL)flag
{
    _isAbstract = flag?YES:NO;
}

- (void) setIsLocal:(BOOL)flag
{
    _isLocal = flag?YES:NO;
}
- (BOOL) isLocal
{
    return _isLocal;
}

- (BOOL) isRootInterface
{
    return self == [IDLInterface rootInterface];
//    return [_name isEqualToString:@"Object"];
}

- initWith:(NSDictionary*)functions inherited:(NSArray*)namesList
{
    self = [super init];
    [self setInheritanceList:namesList];
    [self setFunctionList:functions];
    return self;
}

+ interface:(NSDictionary*)functions inherited:(NSArray*)namesList;
{
    return [[[self allocWithZone:[self zone]]initWith:functions inherited:namesList]autorelease];
}

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter]removeObserver:self];
    [_list release];
    [_inheritance release];
    [_name release];
    [_selectorList release];
    [_methodList release];
    [_connection release];
    [super dealloc];
}

- (NSString*) description
{
    return [NSString stringWithFormat:@"%@<0x%x>",
        NSStringFromClass([self class]), self];
}

- (NSArray*)allKeys;
{
    return [_list allKeys];
}

- (NSString*) name
{
    return _name;
}

- _functionNamed:(NSString*)aName
{
    NSObject* obj = [[self methodList] objectForKey:aName];
    if(!obj){
        NSArray* array = [self inheritanceList];
        NSEnumerator* en = [array objectEnumerator];
        IDLInterface* ifObj;
        while(ifObj = [en nextObject]){
            if(obj = [ifObj _functionNamed:aName])
                return obj;
        }
    }
    return obj;
}

- functionNamed:(NSString*)aName
{
    NSObject* obj = [self _functionNamed:aName];
    if(!obj && ![self isAbstract]){ // abstract is not inherited from Object
        IDLInterface* root = [IDLInterface rootInterface];
        if(self != root)
            return [root _functionNamed:aName];
    }
    return obj;
}

- (NSArray*) allFunctions
{
    NSMutableArray* array = [NSMutableArray array];
    NSEnumerator* en = [_list objectEnumerator];
    NSObject* obj;
    while(obj = [en nextObject]){
        if([obj isKindOfClass:[IDLFunction class]])
            [array addObject:obj];
        else
        if([obj isKindOfClass:[IDLAttribute class]]){
            IDLFunction* fo = [(IDLAttribute*)obj getAttribute];
            if(fo)
                [array addObject:fo];
            if(fo = [(IDLAttribute*)obj setAttribute])
                [array addObject:fo];
        }
    }
    return array;
}

- (NSMutableDictionary*) methodList
{
    NSLock* lock = _lock;
    [lock lock];
    if(!_methodList){
        NSEnumerator* en = [_list objectEnumerator];
        IDLDefinition* obj;
        _methodList = [[NSMutableDictionary alloc]init];
        while(obj = [en nextObject]){
            if([obj isKindOfClass:[IDLFunction class]])
                [_methodList setObject:obj forKey:[obj name]];
            else
            if([obj isKindOfClass:[IDLAttribute class]]){
                IDLFunction* fo = [(IDLAttribute*)obj getAttribute];
                if(fo)
                    [_methodList setObject:fo forKey:[fo name]];
                if(fo = [(IDLAttribute*)obj setAttribute])
                    [_methodList setObject:fo forKey:[fo name]];
            }
        }
    }
    [lock unlock];
    return _methodList;
}

- (IDLAttribute*) _attributeNamed:(NSString*)name
{
	IDLAttribute* attr = [_list objectForKey:name];
	return [attr isKindOfClass:[IDLAttribute class]] ? attr : nil;
}

- (IDLAttribute*) attributeNamed:(NSString*)name
{
	IDLAttribute* attr = [self _attributeNamed:name];
	if(!attr){
		NSEnumerator* en = [[self inheritanceList]objectEnumerator];
		IDLInterface* obj;
		while(obj = [en nextObject])
			if(attr = [obj attributeNamed:name])
				break;
	}
	return attr;
}

- (IDLDefinition*)_definitionNamed:(NSString*)aName
{
    return [_list objectForKey:aName];
}

- (id <IDLDefinition>) localDefinitionNamed:(NSString*)aName
{
    IDLDefinition* obj = [self _definitionNamed:aName];
    if(!obj){
        NSEnumerator* en = [_inheritance objectEnumerator];
        IDLInterface* ifObj;
        while(ifObj = [en nextObject])
            if(obj = (IDLDefinition*)[ifObj localDefinitionNamed:aName])
                break;
    }
    return obj;
}

- (id <IDLDefinition>)definitionNamed:(NSString*)aName
{
    IDLDefinition* obj = (IDLDefinition*)[self localDefinitionNamed:aName];
    if(!obj)
        obj = (IDLDefinition*)[[self parentScope]definitionNamed:aName];
    if(!obj){
        NSEnumerator* en = [_inheritance objectEnumerator];
        IDLInterface* ifObj;
        while(ifObj = [en nextObject])
            if(obj = (IDLDefinition*)[[ifObj parentScope]definitionNamed:aName])
                break;
    }
    return obj;
}

- valueForKey:(NSString*)key
{
	if([key characterAtIndex:0] == '@')
		return [super valueForKey:[key substringFromIndex:1]];
	else
		return [self definitionNamed:key];
}

- (void) addDefinition:(id <IDLDefinition>)def withName:(NSString*)aName
{
    if(!_list)
        _list = [[NSMutableDictionary alloc]init];
    [_list setObject:def forKey:aName];
}

- addDefinition:(id <IDLDefinition>)def withIdentifierArray:(NSArray*)identArray
{
    NSMutableArray* array = [NSMutableArray array];
    id identifier;
    NSEnumerator* en = [identArray objectEnumerator];
    while(identifier = [en nextObject]){
        NSString* dname;
        BOOL isArray = [identifier isKindOfClass:[IDLArrayIdentifier class]];
        if(isArray)
            dname = [identifier name];
        else
            dname = identifier;
                    
        if([self _definitionNamed:dname]){
            NSLog(@"Redefining name '%@'\n",dname); 
        }else{
            IDLTypedef* tdef;
            if(isArray){
                IDLArray* obj = [IDLArray arrayWithType:def identifier:identifier];
                tdef = [IDLTypedef typedefWithType:obj name:dname];
            }else
                tdef = [IDLTypedef typedefWithType:(IDLDefinition*)def name:dname];
            [self addDefinition:tdef withName:dname];
			[array addObject:tdef];
        }
    }
    return [array count] > 0? array : nil;
}

- (id <IDLDefinitionScope>) parentScope
{
    return _parentScope;
}
- (void) setParentScope:(id<IDLDefinitionScope>)scope
{
// FIXME: Should retain parentScope ?
//    [(NSObject*)_parentScope autorelease];
//    _parentScope = [(NSObject*)scope retain];
	_parentScope = scope;
}

- (IDLDefinitionScope*) rootScope
{
    id scope = [self parentScope];
    if(scope)
        return [scope rootScope];
    else
        return [IDLDefinitionScope rootScope];
}

- (NSDictionary*)_definitionDictionary
{
    return _list?_list : [NSMutableDictionary dictionary];
}

- (NSMutableDictionary*)selectorList
{
    NSLock* lock = _lock;
    [lock lock];
    if(!_selectorList){
        NSEnumerator* en = [_list objectEnumerator];
        IDLFunction* obj;
        _selectorList = [[NSMutableDictionary allocWithZone:[self zone]]init];
        while(obj = [en nextObject]){
            if([obj isKindOfClass:[IDLFunction class]]){
                [_selectorList setObject:obj forKey:[obj selectorName]];
            }else
            if([obj isKindOfClass:[IDLAttribute class]]){
                IDLFunction* f = [(IDLAttribute*)obj getAttribute];
                [_selectorList setObject:f forKey:[f selectorName]];
                if(f = [(IDLAttribute*)obj setAttribute])
                    [_selectorList setObject:f forKey:[f selectorName]];
            }
        }
    }
    [lock unlock];
    return _selectorList;
}

- (void) resetSelectors
{
	NSLock* lock = _lock;
    [lock lock];
	if(_selectorList)
		[_selectorList removeAllObjects];
	[lock unlock];
}

- functionWithSelectorNamed:(NSString*)selector
{
    IDLFunction* obj =  [[self selectorList]objectForKey:selector];
    if(!obj){
        NSArray* array = [self inheritanceList];
        NSEnumerator* en = [array objectEnumerator];
        IDLInterface* ifObj;
        while(ifObj = [en nextObject]){
            obj = [ifObj functionWithSelectorNamed:selector];
            if(obj){
                [[self selectorList] setObject:obj forKey:selector]; //cache it here
                return obj;
            }
        }
    }
    return obj;
}

- (void) setNamePath:(NSString*)path
{
    NSEnumerator* en = [_list objectEnumerator];
    IDLDefinition* obj;
    [super setNamePath:path];
    path = path?[NSString stringWithFormat:@"%@/%@", path, [self name]] : [self name];
    while(obj = [en nextObject])
        [obj setNamePath:path];
}

- (BOOL) implementsInterfaceNamed:(NSString*)aName
{
    if([aName isEqualToString:_name] || [aName isEqualToString:@"Object"])
        return YES;
    else{
        NSArray* array = [self inheritanceList];
        NSEnumerator* en = [array objectEnumerator];
        IDLInterface* ifObj;
        while(ifObj = [en nextObject]){
            if([ifObj implementsInterfaceNamed:aName])
                return YES;
        }
    }
    return NO;
}

- (BOOL) adoptsInterface:(IDLInterface*)pifObj
{
    if([[pifObj typeId]isEqualToString:[self typeId]] || ([pifObj isRootInterface] && ![self isAbstract]) ||
            pifObj == [IDLAbstractBase sharedInstance])
        return YES;
    else{
        NSArray* array = [self inheritanceList];
        NSEnumerator* en = [array objectEnumerator];
        IDLInterface* ifObj;
        while(ifObj = [en nextObject]){
            if([ifObj adoptsInterface:pifObj])
                return YES;
        }
    }
    return NO;
}

- (BOOL) _isDerivedFromIDLType:(IDLDefinition*)type
{
	return ([type isKindOfClass:[IDLInterface class]] && [self adoptsInterface:(IDLInterface*)type]) ||
		type == [IDLAbstractBase sharedInstance];
}

// This method is too flexible
- (id <IDLDefinition>) definitionForNamePath:(NSString*)path
{
    if([path length] > 0){
        NSRange r = [path rangeOfString:@"/"];
        if(r.length > 0){
            if(r.location > 0){
                IDLDefinition* obj = (IDLDefinition*)[self definitionNamed:[path substringToIndex:r.location]];
                if([obj respondsToSelector:_cmd])
                    return (IDLDefinition*)[(id<IDLDefinitionScope>)obj definitionForNamePath:[path substringFromIndex:r.location+r.length]];
                else
                    return nil;
            }else
                return [[self rootScope]definitionForNamePath:[path substringFromIndex:r.length]];
        }else
            return [self definitionNamed:path];
    }
    return self;
}

// 'typeId' always contains an absolute path and this method is always sent to the root scope
// 'typeId' is defined ambigiously - first name in the 'path' may be prefix or module
- (id <IDLDefinition>) definitionForTypeId:(NSString*)typeId
{
    return (IDLDefinition*)[[self rootScope] definitionForTypeId:typeId];
}

- (NSNumber*) enumValueForName:(NSString*)n andKey:(NSString*)key 
{
    if(n && key){
        IDLEnum* def = (IDLEnum*)[self definitionForNamePath:n];
        if(def)
            return [def enumValueForKey:key];
        else
            [NSException raise:IDLInvalidArgumentException
                format:@"Cannot locate enum definition for '%@'", n];
    }else
        [NSException raise:IDLInvalidArgumentException
            format:@"NULL argument in [%@ %@]",
            NSStringFromClass([self class]),NSStringFromSelector(_cmd)];
    
    return nil;
}

 - (long) kind
{
    if(_isLocal)
        return tk_local_interface;
    else
    if(_isAbstract)
        return tk_abstract_interface;
    else
        return tk_objref;
}

- (void) _setVersion:(NSString*)version forDefinitionNamed:(NSString*)defName
{
    if(defName){
        defName = [NSString stringWithFormat:@"%@/%@", [self name], defName];
        [(id<IDLScopePrivate>)[self parentScope]_setVersion:version forDefinitionNamed:defName];
    }
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	[super encodeWithCoder:coder];
	if([coder allowsKeyedCoding]){
		[coder encodeObject:_list forKey:@"list"];
		[coder encodeObject:_inheritance forKey:@"inheritance"];
		[coder encodeObject:_name forKey:@"name"];
		[coder encodeConditionalObject:_parentScope forKey:@"parentScope"];
		[coder encodeBool:_isAbstract forKey:@"abstract"];
		[coder encodeBool:_isLocal forKey:@"local"];
	}else{
		[coder encodeObject:_list];
		[coder encodeObject:_inheritance];
		[coder encodeObject:_name];
		[coder encodeConditionalObject:_parentScope];
		[coder encodeValueOfObjCType:@encode(BOOL) at:&_isAbstract];
		[coder encodeValueOfObjCType:@encode(BOOL) at:&_isLocal];
	}
}

- initWithCoder:(NSCoder*)coder
{
	self = [super initWithCoder:coder];
	if([coder allowsKeyedCoding]){
		_list = [[coder decodeObjectForKey:@"list"]retain];
		_inheritance = [[coder decodeObjectForKey:@"inhertance"]retain];
		_name = [[coder decodeObjectForKey:@"name"]retain];
// FIXME: Should retain parentScope ?
//		_parentScope = [[coder decodeObjectForKey:@"parentScope"]retain];
		_parentScope = [coder decodeObjectForKey:@"parentScope"];
		_isAbstract = [coder decodeBoolForKey:@"abstract"];
		_isLocal = [coder decodeBoolForKey:@"local"];
	}else{
		_list = [[coder decodeObject]retain];
		_inheritance = [[coder decodeObject]retain];
		_name = [[coder decodeObject]retain];
// FIXME: Should retain parentScope ?
//		_parentScope = [[coder decodeObject]retain];
		_parentScope = [coder decodeObject];
		[coder decodeValueOfObjCType:@encode(BOOL) at:&_isAbstract];
		[coder decodeValueOfObjCType:@encode(BOOL) at:&_isLocal];
	}
	return self;
}

@end

