/*
 *  AdORB
 *  Copyright (C) 2006 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import <Foundation/Foundation.h>
#import "IDLDefinition_Private.h"
#import "IDLPrimitive.h"
#import "IDLInterface_Private.h"
#import "IDLDefinitionScope_Private.h"
#import "IDLTypedValue.h"
#import "IDLValueType.h"
#import "IDLOctetSequence.h"
#import "IDLOctetArray.h"
#import "IDLArrayIdentifier.h"
#import <expr.tab.h>

@implementation IDLPrimitive

- (int) defType
{
    return _type;
}

- initWithType:(int)type
{
    self = [super init];
    _type = type;
    return self;
}

+ definitionForType:(int) type
{
    return [[[self allocWithZone:[self zone]]initWithType:type]autorelease];
}

-(NSString*) description
{
    return [[super description] stringByAppendingFormat:@"type = %@",stringForType(_type)];
}

- (NSString*) name
{
    NSString* str = stringForType(_type);
    if(str)
        return str;
    else
        return [super name];
}


- (NSString*) valueDescriptionOf:value
{
    if([value isKindOfClass:[IDLTypedValue class]])
        return [(IDLTypedValue*)value valueDescription];
    else
    switch(_type){
        default:
            return [value description];
        case SHORT:
        case INT:
        case LONG:
        case USHORT:
        case UINT:
        case ULONG:
        case UNSIGNED:
            return [value description];
        case BOOLEAN:
            if([value respondsToSelector:@selector(boolValue)])
                return [value boolValue]?@"TRUE":@"FALSE";
            else
                return [value description];
        case STRING:
        case WSTRING:
            if([value isKindOfClass:[NSString class]])
                return [NSString stringWithFormat:@"\"%@\"",value];
            else
                return [value description];
        case CHAR:
            if([value isKindOfClass:[NSNumber class]]){
				unsigned char c = [value unsignedCharValue];
				if(c >= 32 && c < 127)
					return [NSString stringWithFormat:@"\'%c\'",c];
				else
					return [NSString stringWithFormat:@"\'\\%o\'", c];
            }else
                return [NSString stringWithFormat:@"\'%@\'",[value description]];
        case WCHAR:
            if([value isKindOfClass:[NSNumber class]])
				return [NSString stringWithFormat:@"L\'\\u%X\'", [value unsignedShortValue]];
			else
                return [NSString stringWithFormat:@"\'%@\'",[value description]];
        case OCTET:
            if([value isKindOfClass:[NSNumber class]])
                return [NSString stringWithFormat:@"0x%x",[value unsignedCharValue]];
            else
                return [value description];
    }
}

- (BOOL) isOneWay
{
    switch(_type){
        case ONEWAY:
            return YES;
        default:
            return NO;
    }
}

- sequenceClass
{
    switch(_type){
        case OCTET:
            return [IDLOctetSequence class];
        default:
            return [super sequenceClass];
    }
}

- arrayClass
{
    switch(_type){
        case OCTET:
            return [IDLOctetArray class];
        default:
            return [super arrayClass];
    }
}

- (long) kind
{
    switch(_type){
        case CHAR:
            return tk_char;
        case WCHAR:
            return tk_wchar;
        case OCTET:
            return tk_octet;
        case SHORT:
            return tk_short;
        case USHORT:
            return tk_ushort;
        case INT:
            return tk_long;
        case UINT:
            return tk_ulong;
        case LONG:
            return tk_long;
        case ULONG:
            return tk_ulong;
        case UNSIGNED:
            return tk_ulong;
        case LONGLONG:
            return tk_longlong;
        case ULONGLONG:
            return tk_ulonglong;

        case FLOAT:
            return tk_float;
        case DOUBLE:
            return tk_double;
        case FIXED:
            return tk_fixed;
        case BOOLEAN:
            return tk_boolean;
        case STRING:
            return tk_string;
        case WSTRING:
            return tk_wstring;

        case VOID:
            return tk_void;
        case ONEWAY:
            return tk_void;
        default:
            return tk_null;
    }
}

- (void) encodeWithCoder:(NSCoder*)coder
{
	[super encodeWithCoder:coder];
	if([coder allowsKeyedCoding]){
		[coder encodeInt:_type forKey:@"type"];
	}else{
		[coder encodeValueOfObjCType:@encode(int) at:&_type];
	}
}

- initWithCoder:(NSCoder*)coder
{
	self = [super initWithCoder:coder];
	if([coder allowsKeyedCoding]){
		_type = [coder decodeIntForKey:@"type"];
	}else{
		[coder decodeValueOfObjCType:@encode(int) at:&_type];
	}
	return self;
}

@end

#import "IDLAny.h"

NSString* stringForType(int type)
{
    switch(type){
        case CHAR:
            return @"char";
        case WCHAR:
            return @"wchar";
        case OCTET:
            return @"octet";
        case SHORT:
            return @"short";
        case USHORT:
            return @"unsigned short";
        case INT:
            return @"int";
        case UINT:
            return @"unsigned int";
        case LONG:
            return @"long";
        case ULONG:
            return @"unsigned long";
        case UNSIGNED:
            return @"unsigned";
        case LONGLONG:
            return @"long long";
        case ULONGLONG:
            return @"unsigned long long";

        case FLOAT:
            return @"float";
        case DOUBLE:
            return @"double";
        case FIXED:
            return @"fixed";
        case BOOLEAN:
            return @"boolean";
        case STRING:
            return @"string";
        case WSTRING:
            return @"wstring";

        case VOID:
            return @"void";
        case ONEWAY:
            return @"oneway";
        default:
            return nil;
    }
}

void initScope(IDLDefinitionScope* scope)
{
    [scope addDefinition:[IDLPrimitive definitionForType:CHAR] 
                withName:stringForType(CHAR)];
    [scope addDefinition:[IDLPrimitive definitionForType:WCHAR] 
                withName:stringForType(WCHAR)];
    [scope addDefinition:[IDLPrimitive definitionForType:OCTET] 
                withName:stringForType(OCTET)];
    [scope addDefinition:[IDLPrimitive definitionForType:SHORT] 
                withName:stringForType(SHORT)];
    [scope addDefinition:[IDLPrimitive definitionForType:USHORT] 
                withName:stringForType(USHORT)];
    [scope addDefinition:[IDLPrimitive definitionForType:INT] 
                withName:stringForType(INT)];
    [scope addDefinition:[IDLPrimitive definitionForType:UINT] 
                withName:stringForType(UINT)];
    [scope addDefinition:[IDLPrimitive definitionForType:LONG] 
                withName:stringForType(LONG)];
    [scope addDefinition:[IDLPrimitive definitionForType:ULONG] 
                withName:stringForType(ULONG)];
    [scope addDefinition:[IDLPrimitive definitionForType:LONGLONG] 
                withName:stringForType(LONGLONG)];
    [scope addDefinition:[IDLPrimitive definitionForType:ULONGLONG] 
                withName:stringForType(ULONGLONG)];
    [scope addDefinition:[IDLPrimitive definitionForType:UNSIGNED] 
                withName:stringForType(UNSIGNED)];
    [scope addDefinition:[IDLPrimitive definitionForType:FLOAT] 
                withName:stringForType(FLOAT)];
    [scope addDefinition:[IDLPrimitive definitionForType:DOUBLE] 
                withName:stringForType(DOUBLE)];
    [scope addDefinition:[IDLPrimitive definitionForType:FIXED] 
                withName:stringForType(FIXED)];
    [scope addDefinition:[IDLPrimitive definitionForType:BOOLEAN] 
                withName:stringForType(BOOLEAN)];
    [scope addDefinition:[IDLPrimitive definitionForType:STRING] 
                withName:stringForType(STRING)];
    [scope addDefinition:[IDLPrimitive definitionForType:WSTRING] 
                withName:stringForType(WSTRING)];
    [scope addDefinition:[IDLPrimitive definitionForType:VOID] 
                withName:stringForType(VOID)];
    [scope addDefinition:[IDLPrimitive definitionForType:ONEWAY] 
                withName:stringForType(ONEWAY)];
    [scope addDefinition:[IDLAny any]
                withName:@"any"];

}
 
