/*
 *  AdORB
 *  Copyright (C) 2007 Victor Ananiev
 *
 *  This file is part of ADORB.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#import "ORBNotificationCenter.h"
#import "CORBAException.h"
#import "ORB_Private.h"

@class _ORBObserver;

@interface _ORBObserverEntry : NSObject
{
@public
	_ORBObserver* _next;
	NSString* _name;
	id _object;
}
@end
@interface _ORBObserver : _ORBObserverEntry
{
@public
	id _target;
	SEL _selector;
	IMP _method;
}
@end

// _data is a hashtable and contains four kinds of entries (of class _ORBObserverEntry)
// with name != nil and object != nil
// with name != nil and object = nil
// with name = nil and object != nil
// with both name = nil and object = nil
// each entry begins a linked chain of _ORBObserver objects

@implementation ORBNotificationCenter

NSString* ORBWillSendRequest = @"ORBWillSendRequestNotification";
NSString* ORBDidReceiveResponse = @"ORBDidReceiveResponseNotification";
NSString* ORBDidReceiveException = @"ORBDidReceiveExceptionNotification";
NSString* ORBDidReceiveOther = @"ORBDidReceiveOtherNotification";

NSString* ORBDidReceiveRequestContext = @"ORBDidReceiveRequestContextNotification";
NSString* ORBDidReceiveRequest = @"ORBDidReceiveRequestNotification";
NSString* ORBWillSendResponse = @"ORBWillSendResponseNotification";
NSString* ORBWillSendException = @"ORBWillSendExceptionNotification";
NSString* ORBWillSendOther = @"ORBWillSendOtherNotification";

NSString* ORBDidCreateObjectProfile = @"ORBDidCreateObjectProfileNotification";


NSString* ORBServiceContextKey = @"serviceContext";
NSString* ORBOperationNameKey = @"operationName";
NSString* ORBObjectKey = @"object";
NSString* ORBRequestIdKey = @"requestId";
NSString* ORBExceptionKey = @"exception";
NSString* ORBReplyStatusKey = @"replyStatus";
NSString* ORBCompletionStatusKey = @"completed";
NSString* ORBOnewayKey = @"oneway";

+ (void) initialize
{
	if([ORBNotificationCenter class] == self){
	
	}
}

- init
{
	self = [super init];
	_lock = [[NSLock alloc]init];
	_data = [[NSMutableSet alloc]init];
	_set = [[NSMutableSet alloc]initWithObjects:CORBASystemException, CORBAForwardRequest, nil];
	return self;
}

- (void) dealloc
{
	[_lock release];
	[_set release];
	[_data release];
	[super dealloc];
}

- (void)addObserver:anObserver selector:(SEL)aSelector name:(NSString *)notificationName object:anObject
{
	_ORBObserver* observer = [[_ORBObserver alloc]init];
	observer->_name = [notificationName copy];
	observer->_object = anObject;
	observer->_selector = aSelector;
	observer->_target = anObserver;
	if(aSelector && [anObserver respondsToSelector:aSelector])
		observer->_method = [anObserver methodForSelector:aSelector];
	else if(aSelector)
		NSLog(@"[%@ %@] - object %@ does not implement method %@",
			NSStringFromClass([self class]), NSStringFromSelector(_cmd), anObserver, NSStringFromSelector(aSelector));

#if USE_DEFS
	struct { @defs(_ORBObserverEntry) } shadow;
	shadow._name = observer->_name;
	shadow._object = observer->_object;
	shadow.isa = [_ORBObserverEntry class];
#else
	_ORBObserverEntry *shadow = [[_ORBObserverEntry alloc]init];
	shadow->_name = [observer->_name retain];
	shadow->_object = observer->_object;
#endif
	[_lock lock];
#if USE_DEFS
	_ORBObserverEntry * entry = [_data member:(id)&shadow];
#else
	_ORBObserverEntry * entry = [_data member:shadow];
	[shadow release];
#endif
	if(!entry){
		entry = [[_ORBObserverEntry alloc]init];
		entry->_name = [observer->_name retain];
		entry->_object = anObject; // not retained
		[_data addObject:entry];
		[entry release];
	}
	_ORBObserver* last = entry->_next;
	if(last){
		while(last->_next) last = last->_next;
		last->_next = observer;
	}else
		entry->_next = observer;
	
	[_lock unlock];
}

- (void) _postNotification:(NSNotification*)notification toObservers:(NSArray*)array
{
	NSEnumerator* en = [array objectEnumerator];
	_ORBObserver* observer;
	while(observer = [en nextObject]){
		NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
		@try{
			if(observer->_method && observer->_target)
				(*observer->_method)(observer->_target, observer->_selector, notification);
		}@catch(NSException *localException){
			NSLog(@"ORBNotificationCenter: Exception ignored while posting notification '%@', %@",
				[localException name], [localException reason]);
		}
		[pool release];
	}
}

static NSArray* _getObservers(NSSet* data, NSString* name, id object);

- (void)postNotification:(NSNotification *)notification
{
	NSString* name = [notification name];
	id object = [notification object];
	[_lock lock];
	NSArray* array = _getObservers(_data, name, object);
	[_lock unlock];

	[self _postNotification:notification toObservers:array];
}

- (void)postNotificationName:(NSString *)notificationName object:(id)anObject
{
	NSNotification* note = [NSNotification notificationWithName:notificationName object:anObject];
	[self postNotification:note];
}

- (void)postNotificationName:(NSString *)notificationName object:(id)anObject userInfo:(NSDictionary *)userInfo
{
	NSNotification* note = [NSNotification notificationWithName:notificationName 
		object:anObject userInfo:userInfo];
	[self postNotification:note];
}

- (NSException*) processInterceptorStartPost:(NSString*)name object:anObject userInfo:(NSDictionary*)info
		flowStack:(NSMutableArray*)flowStack
{
	NSException *exception = nil;
	NSNotification *notification = [NSNotification notificationWithName:name object:anObject userInfo:info];
	[_lock lock];
	NSArray* array = _getObservers(_data, name, anObject);
	[_lock unlock];
	NSEnumerator* en = [array objectEnumerator];
	_ORBObserver *observer;
	while(exception == nil && (observer = [en nextObject])){
		NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
		@try{
			if(observer->_target && observer->_method)
				(*observer->_method)(observer->_target, observer->_selector, notification);
			[flowStack addObject:observer->_target];
		}@catch(NSException *localException){
			if(![_set containsObject:[localException name]])
				NSLog(@"ORBNotificationCenter: Exception ignored while posting notification '%@', %@",
					[localException name], [localException reason]);
			else
				exception = [localException retain];
		}
		[pool release];
	}
	return exception;
}

- (NSException*) postWillSendRequest:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBWillSendRequest;
	NSException* exception = [self processInterceptorStartPost:name object:anObject userInfo:userInfo flowStack:flowStack];
	if(exception){
		NSMutableDictionary* info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
		[info setObject:exception forKey:@"exception"];
		[exception release];
		if([CORBAException isForwardRequestException:exception]){
			return [self postDidReceiveOther:anObject userInfo:info flowStack:flowStack];
		}else{
			return [self postDidReceiveException:anObject userInfo:info flowStack:flowStack];
		}
	}
	return nil;
}

typedef BOOL (*ExceptionCheck)(NSException* exception);

static BOOL _checkException0(NSException* exception)
{
	return YES;
}
static BOOL _checkException1(NSException* exception)
{
	return [CORBAException isForwardRequestException:exception];
}
static BOOL _checkException2(NSException* exception)
{
	return ![CORBAException isForwardRequestException:exception];
}

- (NSException*) processInterceptorPost:(NSString*)name object:anObject userInfo:(NSMutableDictionary*)info
		flowStack:(NSMutableArray*)flowStack condition:(ExceptionCheck)check
{
	[_lock lock];
	NSArray* array = _getObservers(_data, name, anObject);
	[_lock unlock];
	NSNotification *notification = [NSNotification notificationWithName:name object:anObject userInfo:info];
	id observer;
	BOOL ok = YES;
	while(ok && (observer = [flowStack lastObject])){
		NSAutoreleasePool* pool = [[NSAutoreleasePool alloc]init];
		NSEnumerator* en = [array objectEnumerator];
		_ORBObserver* entry;
		while(ok && (entry = [en nextObject])){
			if(entry->_target == observer && entry->_method){
				@try{
					(*entry->_method)(entry->_target, entry->_selector, notification);
				}@catch(NSException *localException){
					if(![_set containsObject:[localException name]])
						NSLog(@"ORBNotificationCenter: Exception ignored while posting notification '%@', %@",
							[localException name], [localException reason]);
					else{
						[info setObject:localException forKey:@"exception"];
						if((*check)(localException))
							ok = NO;
					}
				}
			}
		}
		[flowStack removeLastObject];
		[pool release];
	}
	return [info objectForKey:@"exception"];
}

- (NSException*) postDidReceiveReply:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBDidReceiveResponse;
	NSMutableDictionary *info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
	NSException* exception = [self processInterceptorPost:name object:anObject userInfo:info flowStack:flowStack condition:&_checkException0];
	if([flowStack count] > 0){
		if([CORBAException isForwardRequestException:exception])
			return [self postDidReceiveOther:anObject userInfo:info flowStack:flowStack];
		else
			return [self postDidReceiveException:anObject userInfo:info flowStack:flowStack];
	}else
		return exception;
}

- (NSException*) postDidReceiveException:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBDidReceiveException;
	NSMutableDictionary *info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
	NSException* exception = [self processInterceptorPost:name object:anObject userInfo:info flowStack:flowStack condition:&_checkException1];
	if([flowStack count] > 0)
		return [self postDidReceiveOther:anObject userInfo:info flowStack:flowStack];
	else
		return exception;
}

- (NSException*) postDidReceiveOther:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBDidReceiveOther;
	NSMutableDictionary *info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
	NSException* exception = [self processInterceptorPost:name object:anObject userInfo:info flowStack:flowStack condition:&_checkException2];
	if([flowStack count] > 0)
		return [self postDidReceiveException:anObject userInfo:info flowStack:flowStack];
	else
		return exception;
}

- (NSException*) postDidReceiveRequestContext:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBDidReceiveRequestContext;
	NSException* exception = [self processInterceptorStartPost:name object:anObject userInfo:userInfo flowStack:flowStack];
	if(exception){
		NSMutableDictionary* info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
		[info setObject:exception forKey:@"exception"];
		[exception release];
		if([CORBAException isForwardRequestException:exception]){
			return [self postWillSendOther:anObject userInfo:info flowStack:flowStack];
		}else{
			return [self postWillSendException:anObject userInfo:info flowStack:flowStack];
		}
	}
	return nil;
}

- (NSException*) postDidReceiveRequest:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBDidReceiveRequest;
	NSException* exception = [self processInterceptorStartPost:name object:anObject userInfo:userInfo flowStack:flowStack];
	if(exception){
		NSMutableDictionary* info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
		[info setObject:exception forKey:@"exception"];
		[exception release];
		if([CORBAException isForwardRequestException:exception]){
			return [self postWillSendOther:anObject userInfo:info flowStack:flowStack];
		}else{
			return [self postWillSendException:anObject userInfo:info flowStack:flowStack];
		}
	}
	return nil;
}

- (NSException*) postWillSendResponse:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBWillSendResponse;
	NSMutableDictionary *info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
	NSException* exception = [self processInterceptorPost:name object:anObject userInfo:info flowStack:flowStack condition:&_checkException0];
	if([flowStack count] > 0){
		if([CORBAException isForwardRequestException:exception])
			return [self postWillSendOther:anObject userInfo:info flowStack:flowStack];
		else
			return [self postWillSendException:anObject userInfo:info flowStack:flowStack];
	}else
		return exception;
}

- (NSException*) postWillSendException:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBWillSendException;
	NSMutableDictionary *info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
	NSException* exception = [self processInterceptorPost:name object:anObject userInfo:info flowStack:flowStack condition:&_checkException1];
	if([flowStack count] > 0)
		return [self postWillSendOther:anObject userInfo:info flowStack:flowStack];
	else
		return exception;
}

- (NSException*) postWillSendOther:anObject userInfo:(NSDictionary*)userInfo flowStack:(NSMutableArray*)flowStack
{
	NSString* name = ORBWillSendOther;
	NSMutableDictionary *info = [NSMutableDictionary dictionaryWithDictionary:userInfo];
	NSException* exception = [self processInterceptorPost:name object:anObject userInfo:info flowStack:flowStack condition:&_checkException2];
	if([flowStack count] > 0)
		return [self postWillSendException:anObject userInfo:info flowStack:flowStack];
	else
		return exception;
}

static void _removeFromEntry(_ORBObserverEntry *entry, id target, NSString* name, id object, BOOL all);

- (void)removeObserver:(id)anObserver
{
	[_lock lock];
	NSEnumerator *en = [_data objectEnumerator];
	_ORBObserverEntry* entry;
	while(entry = [en nextObject])
		_removeFromEntry(entry, anObserver, nil, nil, YES);
	[_lock unlock];
}

- (void)removeObserver:(id)anObserver name:(NSString *)name object:(id)anObject
{
	if(name == nil && anObject == nil)
		[self removeObserver:anObserver];
	else{
		[_lock lock];
		NSEnumerator *en = [_data objectEnumerator];
		_ORBObserverEntry* entry;
		while(entry = [en nextObject])
			_removeFromEntry(entry, anObserver, name, anObject, NO);
		[_lock unlock];
	}
}

- (void) addAllowedExceptionName:(NSString*)name
{
	NSAssert(name != nil, @"name != nil");
	[_set addObject:name];
}

- (void) addAllowedExceptionNamesFromArray:(NSArray*)names
{
	if(names)
		[_set addObjectsFromArray:names];
}

- (void) removeAllowedExceptionName:(NSString*)name
{
	NSAssert(name != nil, @"name != nil");
	[_set removeObject:name];
}

@end

@implementation _ORBObserverEntry

- (unsigned) hash
{
	return [_name hash] ^ (unsigned)_object;
}

- (BOOL) isEqual:(_ORBObserverEntry*)obj
{
	return [obj isMemberOfClass:[_ORBObserverEntry class]] &&
		_object == obj->_object && (_name == obj->_name || _name != nil && [obj->_name isEqualToString:_name]);
}

- (void) dealloc
{
	// _object is not retained
	[_name release];
	[_next release];
	[super dealloc];
}
@end

@implementation _ORBObserver

@end

static void _addObservers(_ORBObserverEntry *entry, NSMutableArray* array, NSMutableArray* objects)
{
	_ORBObserver* observer = entry->_next;
	while(observer){
		[array addObject:observer];
		if(observer->_target) // retain target
			[objects addObject:observer->_target];
		observer = observer->_next;
	}
}

static NSArray* _getObservers(NSSet* data, NSString* name, id object)
{
	NSMutableArray* objects = [NSMutableArray array]; // to retain observers until AutoreleasePool is released
	NSMutableArray* array = [NSMutableArray array];
#if USE_DEFS
	struct { @defs(_ORBObserverEntry) } shadow;
	shadow._name = name;
	shadow._object = object;
	shadow.isa = [_ORBObserverEntry class];
	_ORBObserverEntry* entry = [data member:(id)&shadow];
#else
	_ORBObserverEntry *shadow = [[[_ORBObserverEntry alloc]init]autorelease];
	shadow->_name = [name retain];
	shadow->_object = object;
	_ORBObserverEntry* entry = [data member:shadow];
#endif
	if(entry)
		_addObservers(entry, array, objects);
	if(name != nil){
#if USE_DEFS
		shadow._name = nil;
		if(entry = [data member:(id)&shadow])
			_addObservers(entry, array, objects);
#else
		[shadow->_name release];
		shadow->_name = nil;
		if(entry = [data member:shadow])
			_addObservers(entry, array, objects);
#endif
	}
	if(object != nil){
#if USE_DEFS
		shadow._name = name;
		shandow._object = nil;
		if(entry = [data member:shadow])
			_addObservers(entry, array, objects);
#else
		[shadow->_name release];
		shadow->_name = [name retain];
		shadow->_object = nil;
		if(entry = [data member:shadow])
			_addObservers(entry, array, objects);
#endif
	}
	if(name != nil && object != nil){
#if USE_DEFS
		shadow._name = nil;
		shadow._object = nil;
		if(entry = [data member:(id)&shadow])
			_addObservers(entry, array, objects);
#else		
		[shadow->_name release];
		shadow->_name = nil;
		shadow->_object = nil;
		if(entry = [data member:shadow])
			_addObservers(entry, array, objects);
#endif
	}
	return array;
}

static void _removeFromEntry(_ORBObserverEntry *entry, id target, NSString* name, id object, BOOL all)
{
	_ORBObserver* observer = entry->_next;
	while(observer){
		if(observer->_target == target && 
			(all || ((object == nil || observer->_object == object) && (name == nil || [observer->_name isEqualToString:name]))))
		{
			entry->_next = observer->_next;
			[observer release];
			observer = entry->_next;
		}else{
			entry = observer;
			observer = observer->_next;
		}
	}
}

